@@ -4,84 +4,282 @@ Created on Mon Jan  6 21:16:58 2020
 
 @author: skjerns
 """
+import importlib
 import os, sys
+from os.path import join
+from os import access
 import numpy as np
 import logging
-from .utils import validkey, strip, lowercase
+from .utils import validkey, strip, lowercase, make_key, valid_filename
 from .utils import read_csv, write_csv
 from xml.etree import ElementTree as ET
 from xml.etree.ElementTree import Element
+from copy import deepcopy
+
+try: profile #a hack for not having to remove the profile tags when not testing
+except NameError: profile = lambda x: x   # pass-through decorator
+
+def get_module(name):
+    try:
+        module = importlib.import_module(name)
+        return module
+    except ModuleNotFoundError:
+        print(f'{name} is not installed. Install with pip install {name}')
+        return False
+    raise Exception(f'Cant load module {name}')
+    
 
 class Entry():
     """
-    Hello, it's me, Entry
+    Base class for Unisens entries. All other entries inherit from this.
     
-    :param test:test
+    An Entry has an .attrib dict that specifies the attributes.
+    It can have a parent object and child-entries.
+    Attributes can be set via Entry.set_attrib(name, value) or directly
+    via setting instance.name = value.
+    Children can be accessed via instance['child'] or instance.child
+
+    Parameters
+    ----------
+    attrib : dict, optional
+        DESCRIPTION. an attribute dictionary  e.g. {duration:'4'}
+    parent : Entry, optional
+        DESCRIPTION. a folder or a Entry type.
+    **kwargs : TYPE
+        DESCRIPTION.
     """
+    
+
     def __len__(self):
         return len(self._entries)
-    
+
     def __iter__(self):
-        return self._entries.__iter__()
-    
+        return list(self._entries).__iter__()
+
     def __repr__(self):
         return "<{}({})>".format(self._name, self.attrib)
  
+    # @profile
     def __init__(self, attrib=None, parent='.', **kwargs):
-        if attrib is None: 
+
+        if attrib is None:
             attrib=dict()
-        self.attrib = attrib
+        self.__dict__['attrib'] = attrib
         self.__dict__.update(self.attrib)
-        self._entries = list()
-        self._folder = parent._folder if isinstance(parent, Entry) else parent
-        self._parent = parent if isinstance(parent, Entry) else None
-        self._name = lowercase(type(self).__name__)
+        self.__dict__['_entries'] = []
+        self.__dict__['_folder'] = parent.__dict__['_folder'] if isinstance(parent, Entry) else parent
+        self.__dict__['_parent'] = parent if isinstance(parent, Entry) else None
+        self.__dict__['_name'] = lowercase(type(self).__name__)
         for key in kwargs:
             self.key = kwargs[key]
         self._autosave()
-    
-    
+
+    # @profile
+    def __contains__(self, item):
+        if item in self.__dict__: return True
+        if make_key(item) in self.__dict__: return True
+        try:
+            self.__getitem__(item)
+            return True
+        except:
+            return False
+
+    # @profile
     def __setattr__(self, name:str, value:str):
         """
-        This will allow to set new attributes with .attrname = value
-        while warning the user if builtin methods are overwritten
+        Allows settings of attributes via .name = value.
         """
-        methods = dir(type(self))
         super.__setattr__(self, name, value)
+        if name.startswith('_'): return
+        methods = dir(type(self))
         # do not overwrite if it's a builtin method
-        if name not in methods and not name.startswith('_') and \
+        if name not in methods and   \
             isinstance(value, (int, float, bool, bytes, str)):
             self.set_attrib(name, value)
 
+    # @profile
+    def __getattr__(self, key):
+        if key == "__setstate__":
+            raise AttributeError(key)
+        try: return self.__dict__[key]
+        except: pass
+        try:
+            i, key2 = self._get_index(key)
+            return self.__dict__[key2]
+        except KeyError:  
+            return self.__getattribute__(key)
+        raise AttributeError(f'{key} not found')
+
+    # @profile
+    def __getitem__(self, key):
+        if isinstance(key, str):
+            i, key = self._get_index(key)
+            return self.__dict__[key]
+        elif isinstance(key, int):
+            return self._entries[key]
+        raise KeyError(f'{key} not found')
+
     def _autosave(self):
         """
-        if autosave is enabled, this function will call the autosave
-        function of parents until the Unisens object is reached and
-        then save when anything is changed.
+        if autosave is enabled, this function will call the autosave function
+        of parents until the uppermost Unisens object is reached and then
+        save when anything is changed.
         """
         try:
             if self._parent is not None:
                 self._parent._autosave()
-        except:
+        except: # there might an error, then do nothing.
             pass
 
-    def add_entry(self, entry:'Entry'):
+    def _check_readonly(self):
+        """
+        will raise an exception if a write operation 
+        is requested to readonly file
+        """
+        if hasattr(self, '_parent') and self._parent is not None:
+            if isinstance(self._parent, str): return True
+            self._parent._check_readonly()
+        elif hasattr(self, '_readonly'):
+            if self._readonly:
+                raise IOError(f'Read only, can\'t write to {self._folder}.')
+        return True
+
+    # @profile
+    def _get_index(self, id_or_name, raises=True):
+        """
+        Receive the index and key-name of an object.
         
+        Sub-entries are saved at ._entries as objects in a list and
+        in the __dict__ of the parent. The name of the object is either
+        the id (eg. samples.csv), if it is a FileEntry, or the name 
+        (eg. binFileFormat). This name is made into a attribute-accessible 
+        key by make_key(), e.g. samples_csv. This function searches
+        for the entry in ._entry and returns its index as well as its
+        key name in the __dict__
+
+        Parameters
+        ----------
+        id_or_name : str
+            The name or id of the Entry.
+        raises : bool, optional
+            Raise an exception if not found. The default is True.
+
+        Returns
+        -------
+        [index in ._entries, key-name in __dict__].
+        """
+        
+        id_or_name = id_or_name.upper()
+        id_or_name_key = make_key(id_or_name)
+        
+        # we don't care about case, gently ignoring Linux file case-sensitivity
+        # first check for exact match
+        for i, entry in enumerate(self._entries):
+            if hasattr(entry, 'id'):
+                id = entry.id
+                id_key = make_key(id)
+                if id_or_name.upper()==id:   
+                    return i, id_key # check for exact match
+                if id_key.upper()==id_or_name_key: 
+                    return i, id_key # check for match in key notation
+            else:
+                name = entry._name
+                name_key = make_key(name)
+                if name.upper()==id_or_name.upper(): # same as above
+                    return i, name
+                if name_key.upper()==id_or_name_key: 
+                    return i, name
+
+        found = []
+        for i, entry in enumerate(self._entries):
+            if hasattr(entry, 'id'):
+                id = entry.id.replace('\\', '/').upper() # normalize to linux-slash
+                no_ext = id.rsplit('.', 1)[0] # remove file extension
+                # check if file without extension was requested
+                # e.g. 'test' for test.txt
+                if no_ext==id_or_name:
+                    found+=[(i, make_key(entry.id))]
+                # contains a slash/subdir, so we need to trim that off
+                elif '/' in id:
+                    # e.g. 'test' was requested for 'sub/test.txt'
+                    if os.path.basename(no_ext)==id_or_name:
+                        found+=[(i, make_key(entry.id))]
+                    # e.g. 'test.txt' was requested for 'sub/test.txt'
+                    elif os.path.basename(id)==id_or_name:
+                        found+=[(i, make_key(entry.id))]
+                # no subdir, but full match
+                elif id==id_or_name:
+                    found+=[(i, make_key(entry.id))]
+
+
+
+        if len(found)==1: return found[0]
+        if len(found)>1: raise IndexError(f'More than one match for {id_or_name}: {found}')
+        raise KeyError(f'{id_or_name} not found')
+        
+    def copy(self):
+        """
+        Create a deep copy of this Entry.
+        
+        All references to the parent Unisens object will be removed before.
+        
+        Returns
+        -------
+        copy : Entry
+            An Entry object with all attributes as the invoking object.
+
+        """
+        if hasattr(self, '_parent'):
+            _parent = self._parent
+            del self._parent
+            copy = deepcopy(self)
+            self._parent = _parent
+        else:
+            copy = deepcopy(self)
+        return copy
+
+    # @profile
+    def add_entry(self, entry:'Entry', stack:bool=True):
+        """
+        Add an subentry to this entry
+        
+        The Entry will be added to the ._entries list and
+        to the .__dict__ with its name or id as a key.
+        The subentry can be accessed via entry.entry_name, entry['entry_name']
+        or entry._entries[index]
+
+        Parameters
+        ----------
+        entry : Entry
+            a subentry that should be added.
+        stack : bool, optional
+            If True, an Entry with the same name/id will be added into a list.
+            If False, an already existing Entry will be overwritten.
+            The default is True.
+
+        Returns
+        -------
+        TYPE
+            DESCRIPTION.
+        """
         # there are several Entries that have reserved names.
         # these should not exist double, therefore they are re-set here
         reserved = ['binFileFormat', 'csvFileFormat', 'customFileFormat']
-        for name in reserved:
-            if entry._name==name and name in self.__dict__:
-                if isinstance(self.csvFileFormat, Entry): 
-                    self.remove_entry(name)
+        if entry._name in reserved: 
+            stack=False
         
+        name = entry.attrib.get('id', entry.__dict__['_name'])
+        
+        if not stack:
+            try: self.remove_entry(name)
+            except: pass
         
         self._entries.append(entry)
-        name = entry._name
-        
+                
         # if an entry already exists with this exact name
         # we put the entry inside of a list and append the new entry
         # with the same name. This way all entries are saved
+        name = make_key(name)
         if name in self.__dict__:
             if isinstance(self.__dict__[name], list):
                 self.__dict__[name].append(entry)
@@ -89,38 +287,34 @@ class Entry():
                 self.__dict__[name] = [self.__dict__[name]]
                 self.__dict__[name].append(entry)
         else:
-            self.__dict__[entry._name] = entry
+            self.__dict__[name] = entry
+        entry.__dict__['_parent'] = self
         self._autosave()
         return self
-
-
+    
+    # @profile
     def remove_entry(self, name):
-        deleted = False
-        for i, entry in enumerate(self._entries):
-            if entry._name == name:
-                del self._entries[i]
-                try: del self.__dict__[name]
-                except: pass
-                deleted = True
-            if hasattr(entry, 'id') and entry.id==name:
-                del self._entries[i]
-                try: del self.__dict__[name]
-                except: pass
-                deleted = True
-            if deleted: 
-                self._autosave()
-                return
-        if name in self.__dict__:
-            del self.__dict__[name]
-            deleted = True
-            
-        if deleted: 
-            self._autosave()
-            return
-        
-        raise Exception(f'cannot find entry {name}')
+        """
+        Removes an subentry by name.
 
+        Parameters
+        ----------
+        name : str
+            the name of the entry. 
+            Can be abbreviated, e.g. 'samples' instead of 'samples.csv'.
 
+        """
+        i, key = self._get_index(name)
+        entry = self._entries[i]
+        del self._entries[i]
+        del self.__dict__[key]
+        # if this is an unisens object, also delete the referer there
+        if self.__dict__.get('entries') is not None:
+            for key, e in list(self.__dict__['entries'].items()):
+                if e==entry: del self.__dict__['entries'][key]
+        return self
+
+    # @profile
     def set_attrib(self, name:str, value:str):
         """
         Set an attribute of this Entry
@@ -130,26 +324,42 @@ class Entry():
         name : str
             The name of this attribute
         value : (str, int, float)
-            DESCRIPTION.
-
-        Returns
-        -------
-        self
-
+            value to be added. will be converted to string.
         """
-        name = str(name)
-        value = str(value)
         name = validkey(name)
         self.attrib[name] = value
         self.__dict__.update(self.attrib)
         self._autosave()
         return self
-    
+
+    def get_attrib(self, name:str, default=None):
+        """
+        Retrieves an attribute of this Entry
+
+        Parameters
+        ----------
+        name : str
+            The name to be retrieved.
+        default : str, optional
+            What should be returned if not present. The default is None.
+
+        """
+
+        return self.attrib.get(name, default)
+
     def remove_attr(self, name:str):
         """
         Removes a custom attribute/value of this entry.
-        
-        :param key: The name of the custom attribute
+
+        Parameters
+        ----------
+        name : str
+            The name of the custom attribute.
+
+        Returns
+        -------
+        TYPE
+            DESCRIPTION.
         """
         if name in self.attrib: 
             del self.attrib[name]
@@ -159,20 +369,47 @@ class Entry():
         self._autosave()
         return self           
         
-
     def to_element(self):
-        element = Element(self._name, attrib=self.attrib.copy())
+        """
+        Converts this Entry and all its subentries into an XML Element.
+
+        Returns
+        -------
+        element : Element
+            An xml.etree.Element instance to be written to xml
+
+        """
+        attrib = {}
+        for key, value in self.attrib.items():
+            attrib[key] = str(value)
+        element = Element(self._name, attrib=attrib.copy())
         element.tail = '\n'
         for subelement in self._entries:
             element.append(subelement.to_element())
         return element
     
-    
     def to_xml(self):
+        """
+        Creates a string representing this Entry and all its sub-entries
+        as XML.
+
+        Returns
+        -------
+        str
+            XML string representing this Entry instance.
+
+        """
+        
         element = self.to_element()
         return ET.tostring(element).decode()
     
-    
+    def print_summary(self, indent=0):
+        """
+        Prints a summary of this object containing all entries
+        """
+        print('\t'*indent, self)
+        for entry in self._entries:
+            entry.print_summary(indent+1)
 
 class FileEntry(Entry):
     """
@@ -184,22 +421,35 @@ class FileEntry(Entry):
     def __repr__(self):
         id = self.attrib.get('id', 'None')
         return "<{}({})>".format(self._name, id)
-    
+
+    # @profile
     def __init__(self, id, attrib=None, parent='.', **kwargs):
+        if isinstance(id, str): id = id.replace('\\', '/')
         super().__init__(attrib=attrib, parent=parent, **kwargs)
         if 'id' in self.attrib:
-            self._filename = os.path.join(self._folder, self.id)
-            if not os.path.exists(self._filename):
-                logging.error('File for {} does not exist'.format(self.id))     
+            self._filename = self._folder + '/' +  self.id
+            if not os.access(self._filename, os.F_OK):
+                logging.error('File {} does not exist'.format(self.id))
+                folder = os.path.dirname(self._filename)
+                if not os.path.exists(folder):
+                    os.makedirs(folder, exist_ok=True)
             self.set_attrib('id', self.attrib['id'])
         elif id:
             if os.path.splitext(str(id))[-1]=='':
-                logging.warning('id should be a filename, ie. .bin/.csv/...')
+                logging.warning('id should be a filename with extension ie. .bin')
             self._filename = os.path.join(self._folder, id)
             self.set_attrib('id', id)
+            folder = os.path.dirname(self._filename)
+            if not os.path.exists(folder):
+                os.makedirs(folder, exist_ok=True)
         else:
             raise ValueError('id must be supplied')
+        valid_filename(self.id)
         if isinstance(parent, Entry): parent.add_entry(self)
+        
+        
+
+
 
         
 class SignalEntry(FileEntry):
@@ -207,51 +457,75 @@ class SignalEntry(FileEntry):
     def __init__(self, id=None,  attrib=None, parent='.', **kwargs):
         super().__init__(id=id, attrib=attrib, parent=parent, **kwargs)
         
-        
     def get_data(self, scaled:bool=True, return_type:str='numpy') -> np.array:
         """
         Will try to load the binary data using numpy.
         This might not always work as endianess can't be determined
         with numpy
-        
-        :param scaled: Scale values using scaling factor or return raw numbers
-        :param return_type: numpy, list or pandas. Pandas row indices have the 
-                            names of the corresponding channels. List will be
-                            in format [[ch_name1, data],[ch_name2, data]]
-                            numpy will be as plain numpy array without ch_names
+
+        Parameters
+        ----------
+        scaled : bool, optional
+            Scale values using lsb factor or return raw numbers.
+            The default is True.
+
+        Returns
+        -------
+        np.ndarray
+            The loaded binary data, in this case as numpy array.
+
         """
-        if return_type!='numpy': raise NotImplementedError
+
         n_channels = len(self.channel) if isinstance(self.channel, list) else 1
         dtypestr = self.dataType.lower()
         dtype = np.__dict__.get(dtypestr, f'UNKOWN_DATATYPE: {dtypestr}')
         data = np.fromfile(self._filename, dtype=dtype)
         if scaled:
             data = (data * float(self.lsbValue))
-        data = data.reshape([n_channels, -1])
-        return data
+        return data.reshape([-1, n_channels]).T
     
     
     def set_data(self, data:np.ndarray, dataType:str=None, ch_names:list=None, 
-                       sampleRate:int=None, lsbValue:float=1, unit:str=None, 
+                       sampleRate:int=256, lsbValue:float=1, unit:str=None, 
                        comment:str=None, contentClass:str=None,
-                       adcResolution:int=str):
+                       adcResolution:int=None, baseline:float=None, **kwargs):
         """
         Set the data that is connected to this SignalEntry.
         The data will in any case be saved with Endianness LITTLE,
         as this is the default for numpy. Data will be saved using
         numpy binary data output.
-        
-        :param data: an numpy array
-        :param dataType: the data type of the data. If None, is array.dtype.
-                         Will be used for binary stream output format.
-        :param sampleRate: the sample rate of this data
-        :param lsbValue: the value with which this data is scaled.
-                         this means that       
-        :param comment: sets a comment associated with this data
-        :param unit: the unit which is indicated, e.g. mV, Ohm, ...
-        :param contentClass: sets the content class, e.g. EEG, ECG, ...
+
+        Parameters
+        ----------
+        data : np.ndarray
+            an numpy array.
+        dataType : str, optional
+            The data type of the data. If None, is infered automatically.
+            Can be 'DOUBLE', 'FLOAT', 'INT16', 'INT32', 
+                   'INT8', 'UINT16', 'UINT32', 'UINT8'
+            Will be used for binary stream output format. The default is None.
+        ch_names : list, optional
+            DESCRIPTION. The default is None.
+        sampleRate : int, optional
+            the sample rate of this data. The default is 256.
+        lsbValue : float, optional
+            the value with which this data is scaled this means that. 
+            The default is 1.
+        unit : str, optional
+            the unit which is indicated, e.g. mV, Ohm. The default is None.
+        comment : str, optional
+            sets a comment associated with this data. The default is None.
+        contentClass : str, optional
+            sets the content class, e.g. EEG, ECG. The default is None.
+        adcResolution : int, optional
+            DESCRIPTION. The default is None.
+        baseline : float, optional
+            DESCRIPTION. The default is None.
+        **kwargs : TYPE
+            DESCRIPTION.
         """
-    
+        self._check_readonly()
+
         file = os.path.join(self._folder, self.id)
         
         # if list, convert to numpy array
@@ -259,20 +533,26 @@ class SignalEntry(FileEntry):
             # if the list entries are arrays, we can infer the dtype from them
             if isinstance(data[0], np.ndarray) and dataType is None:
                 dtype = str(data[0].dtype)
+            else:
+                dtype = type(data[0][0])
+            if dtype==np.int64 or dtype==int: dtype=np.int32
             data = np.array(data, dtype=dtype)
-
+        
+        data = np.atleast_2d(data)
+        order = sys.byteorder.upper() # endianess
+        fileFormat = MiscEntry('binFileFormat', key='endianess', value=order)
+        self.add_entry(fileFormat)
 
         #### dtype inference start
-        dtype_mapping = {'float32': 'float',
+        dtype_mapping = {'float16': 'float',
+                         'float32': 'float',
                          'float64': 'double',
                          'int': 'int32'}
 
         if dataType is None:
             dataType = str(data.dtype).upper()
 
-
         dataType = dtype_mapping.get(dataType.lower(), dataType)
-        # infer dtype from array if not indicated
         allowed_dtypes = ['DOUBLE', 'FLOAT', 'INT16', 'INT32', 
                           'INT8', 'UINT16', 'UINT32', 'UINT8']
         assert dataType.upper() in allowed_dtypes,\
@@ -291,7 +571,7 @@ class SignalEntry(FileEntry):
             # this means new channel names are indicated and will overwrite.
             assert len(ch_names)==len(data), f'len {ch_names}!={len(data)}'
             if hasattr(self, 'channel'):
-                logging.warn('Channels present will be overwritten')
+                logging.warning('Channels present will be overwritten')
                 self.remove_entry('channel')
             for name in ch_names:
                 channel = MiscEntry('channel', key='name', value=name)
@@ -306,65 +586,70 @@ class SignalEntry(FileEntry):
         else:
             # this means there are channel names there but do not match n_data
             raise ValueError('Must indicate channel names')
-            
-        # save data using numpy 
-        data.tofile(file)
         
-        # add the file format description (dont understand why dtype isnt here)
-        order = sys.byteorder.upper() # endianess
-        fileFormat = MiscEntry('binFileFormat', key='endianess', value=order)
-        self.add_entry(fileFormat)
+        # save data using numpy , 
+        # .T because unisens reads rows*columns not columns*rows like numpy
+        data.T.tofile(file)
         
+        if baseline is not None: self.set_attrib('baseline', baseline)
         if sampleRate is not None: self.set_attrib('sampleRate', sampleRate)
         if lsbValue is not None: self.set_attrib('lsbValue', lsbValue)
         if unit is not None: self.set_attrib('unit', unit)
         if comment is not None: self.set_attrib('comment', comment)
         if contentClass is not None: self.set_attrib('contentClass', contentClass)
-        if dataType is not None: self.set_attrib('dataType', dataType)
+        if dataType is not None: self.set_attrib('dataType', dataType.lower())
         if adcResolution is not None: self.set_attrib('adcResolution',adcResolution)
+        
+        # set all other keyword arguments/comments as well.
+        for key in kwargs:
+            self.set_attrib(key, kwargs[key])
+            
         self._autosave()
         return self        
 
 class CsvFileEntry(FileEntry):
-    
+    """
+    A FileEntry that links a csv file.
+    """
     def __init__(self, id=None, attrib=None, parent='.', 
                  decimalSeparator='.', separator=';', **kwargs):
         super().__init__(id=id, attrib=attrib, parent=parent, **kwargs)
+        assert decimalSeparator and separator, 'Must supply separators'
 
         if not self.id.endswith('csv'):
-            logging.warn(f'id "{id}" does not end in .csv')
+            logging.warning(f'id "{id}" does not end in .csv')
+        
+        csvFileFormat = MiscEntry('csvFileFormat', parent=self)
+        csvFileFormat.set_attrib('decimalSeparator', decimalSeparator)
+        csvFileFormat.set_attrib('separator', separator)
+        self.add_entry(csvFileFormat)
         
-        if decimalSeparator is not None and separator is not None:
-            csvFileFormat = MiscEntry('csvFileFormat', parent=self)
-            csvFileFormat.set_attrib('decimalSeparator', decimalSeparator)
-            csvFileFormat.set_attrib('separator', separator)
-            self.add_entry(csvFileFormat)
             
-        if not 'csvFileFormat' in self.__dict__:
-            logging.info(f'csvFileFormat information missing for {self.id},'\
-                         ' assuming decimal=. and seperator=;')
-            csvFileFormat = MiscEntry(name = 'csvFileFormat', parent=self)
-            csvFileFormat.set_attrib('decimalSeparator', '.')
-            csvFileFormat.set_attrib('separator', ';')
-            self.add_entry(csvFileFormat)
     
     def set_data(self, data:list, **kwargs):
         """
-        Set data to this csv object.
-        
-        :param data: the data as list or np array
-        :param comment: a comment, that will be added to the first line with #
+        Set data of this csv object.
+
+
+        Parameters
+        ----------
+        data : list
+            the data as list or np array.
+        **kwargs : str
+            DESCRIPTION.
         """
-        
+    
+        self._check_readonly()
+
         assert 'csvFileFormat' in self.__dict__, 'csvFileFormat information'\
                                     'missing: No separator and decimal set'
-        assert isinstance(data, (list,np.ndarray)),'data must be list of lists'
-        
+        assert isinstance(data, (list, np.ndarray)),\
+            f'data must be list of lists, is {type(data)}'
         sep = self.csvFileFormat.separator
         dec = self.csvFileFormat.decimalSeparator
-        
-        n_cols = len(data[0])
-        if n_cols<2: logging.warn('Should supply at least two columns: '\
+
+
+        if len(data)==0 or len(data[0])<2: logging.warning('Should supply at least two columns: '\
                                   'time and data')
         
         write_csv(self._filename, data, sep=sep, decimal_sep=dec)
@@ -401,6 +686,34 @@ class CsvFileEntry(FileEntry):
         return lines
 
 
+    def get_times(self):
+        """
+        Retrieves the times or samples of this CSV entry
+
+        Returns
+        -------
+        times
+            list of int or float denotating the times of the events
+        """
+        data = self.get_data()
+        times, _ = zip(*data)
+        return list(times)
+        
+    
+    def get_labels(self):
+        """
+        Retrieves the labels this CSV entry
+
+        Returns
+        -------
+        times
+            list of int or float denotating the times of the events
+        """
+        data = self.get_data()
+        _, labels = zip(*data)
+        return list(labels)
+
+
 class ValuesEntry(CsvFileEntry):
     """
     Hello, it's me, ValuesEntry
@@ -426,7 +739,7 @@ class ValuesEntry(CsvFileEntry):
             # this means new channel names are indicated and will overwrite.
             assert len(ch_names)==n_cols, f'len {ch_names}!={len(data)}'
             if hasattr(self, 'channel'):
-                logging.warn('Channels present will be overwritten')
+                logging.warning('Channels present will be overwritten')
                 self.remove_entry('channel')
             for name in ch_names:
                 channel = MiscEntry('channel', key='name', value=name)
@@ -455,7 +768,7 @@ class CustomEntry(FileEntry):
     def __init__(self, id=None, **kwargs):
         super().__init__(id=id, **kwargs)  
         self._autosave()
-        
+
     def get_data(self, dtype='auto'):
         """
         Will load the binary data of this CustomEntry.
@@ -465,24 +778,22 @@ class CustomEntry(FileEntry):
             image: .jpeg .jpg .bmp .png .tif .gif
             json:  .json (using json-tricks or json)
             numpy: .npy
+            pickle: .pkl
             binary: anything else
         
         :param dtype: [binary, image, text, numpy, json]
         :returns: the binary data or the otherwise loaded data
         """
-        try:
-            import json_tricks as json
-        except:
-            import json
-            logging.warn('json_tricks not installed, can\'t load numpy array '\
-                         'automatically. install with pip install json-tricks')
 
         if dtype=='auto':
+            
             ext = os.path.splitext(self._filename)[-1].lower()
             txt_exts = ['.txt', '.csv', '.ini']
             img_exts = ['.jpeg', '.jpg', '.bmp', '.png', '.tif', '.gif']
             
-            if ext in img_exts:
+            if hasattr(self, 'dataType'): 
+                dtype = self.dataType
+            elif ext in img_exts:
                 dtype='image'
             elif ext in txt_exts:
                 dtype='text'
@@ -490,26 +801,32 @@ class CustomEntry(FileEntry):
                 dtype='json'
             elif ext=='.npy':
                 dtype='numpy'
+            elif ext=='.pkl':
+                dtype='pickle'
             else:
                 dtype='binary'
                 
         if dtype=='binary':
             with open(self._filename, 'rb') as f:
                 data = f.read()
+        elif dtype=='csv':
+            data = read_csv(self._filename)
         elif dtype=='text':
             with open(self._filename, 'r') as f:
                 data = f.read()
         elif dtype=='image':
-            try: 
-                import imageio
-                data = imageio.imread(self._filename)
-            except: 
-                logging.error('can\'t load: imageio not installed. \n'\
-                                  'run pip install imageio')
-                with open(self._filename, 'rb') as f:
-                    data = f.read()
-            
+            imageio = get_module('imageio')
+            data = imageio.imread(self._filename)
+
+        elif dtype=='pickle':
+            pickle = get_module('pickle')
+            with open(self._filename, 'rb') as f:
+                data = pickle.load(f)
         elif dtype=='json':
+            try:
+                import json_tricks as json
+            except:
+                json = get_module('json')
             with open(self._filename, 'r') as f:
                 data = json.load(f)
         elif dtype=='numpy':
@@ -517,9 +834,9 @@ class CustomEntry(FileEntry):
         else:
             raise ValueError('unknown dtype {}'.format(dtype))
             
+        self.dataType = dtype
         return data
     
-    
     def set_data(self, data, dtype='auto', **kwargs):
         """
         Will save custom data to disk.
@@ -528,18 +845,12 @@ class CustomEntry(FileEntry):
         :param dtype: binary or image.
         :returns: the binary data or an PIL.Image
         """
-        try:
-            import json_tricks as json
-            tricks_installed = True
-        except:
-            import json
-            tricks_installed = False
-            logging.warn('json_tricks not installed, can\'t save numpy array'\
-                         'automatically. install with pip install json-tricks')
-        
+        self._check_readonly()
+
+        # infer datatype automatically
         if dtype=='auto':
             ext = os.path.splitext(self._filename)[-1].lower()
-            txt_exts = ['.txt', '.csv', '.ini']
+            txt_exts = ['.txt', '.ini', '.csv']
             img_exts = ['.jpeg', '.jpg', '.bmp', '.png', '.tif', '.gif']
             
             if ext in img_exts:
@@ -548,11 +859,14 @@ class CustomEntry(FileEntry):
                 dtype='text'
             elif ext=='.json':
                 dtype='json'
+            elif ext=='.pkl':
+                dtype='pickle'
             elif ext=='.npy':
                 dtype='numpy'
             else:
                 dtype='binary'
-                
+        
+        # file saving from here on
         if dtype=='binary':
             with open(self._filename, 'wb') as f:
                 f.write(data)
@@ -560,11 +874,19 @@ class CustomEntry(FileEntry):
             with open(self._filename, 'w') as f:
                 f.write(data)
         elif dtype=='image':
-            try: import imageio
-            except: logging.error('can\'t save: imageio not installed. \n'\
-                                  'run pip install imageio')
+            imageio = get_module('imageio')
             imageio.imsave(self._filename, data)
+        elif dtype=='pickle':
+            pickle = get_module('pickle')
+            with open(self._filename, 'wb') as f:
+                data = pickle.dump(data, f, protocol=3)
         elif dtype=='json':
+            try:
+                import json_tricks as json
+                tricks_installed = True
+            except:
+                json = get_module('json')
+                tricks_installed = False   
             with open(self._filename, 'w') as f:
                 if tricks_installed:
                     json.dump(data, f, allow_nan=True)
@@ -577,6 +899,10 @@ class CustomEntry(FileEntry):
             
         for key in kwargs:
             self.set_attrib(key, kwargs[key])
+            
+        # save dtype within the entry
+        self.dataType = dtype
+        
         self._autosave()
         return self
     
@@ -615,6 +941,7 @@ class MiscEntry(Entry):
             self.set_attrib(key, value)
         self._autosave()
         
+        
 class CustomAttribute(MiscEntry):
     def __new__(*args,**kwargs):
         return MiscEntry('customAttribute', **kwargs)
